/* 
 * $Id: ripconf.c,v 1.3 1998/11/23 22:46:31 masaki Exp $
 */

#include <mrt.h>
#include <rip.h>
#include <config_file.h>
#include <protoconf.h>

/* Comments...
 * Config can work from a file or from user configuration via a
 * socket.
 */

int rip2_show (uii_connection_t *uii) { 
	return (rip_show (PROTO_RIP, uii)); }
int rip2_show_routing_table (uii_connection_t *uii) {
	return (rip_show_routing_table (PROTO_RIP, uii)); }
#ifdef HAVE_IPV6
int ripng_show (uii_connection_t *uii) { 
	return (rip_show (PROTO_RIPNG, uii)); }
int ripng_show_routing_table (uii_connection_t *uii) {
	return (rip_show_routing_table (PROTO_RIPNG, uii)); }
#endif /* HAVE_IPV6 */


typedef struct _dlist_t {
    int num;
    char *in_or_out;
    interface_t *interface;
} dlist_t;


static int
config_distribute_list (uii_connection_t * uii,
			int num, char *in_or_out, int opt_num, char *name)
{
    interface_t *interface;
    dlist_t *dlist;
    rip_t *rip = RIP;

#ifdef HAVE_IPV6
    if (CONFIG_MRTD->protocol == PROTO_RIPNG)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    if (num < 0 || num >= MAX_ALIST) {
	config_notice (TR_ERROR, uii,
		       "wrong access-list number %d (should be 0 <= x < %d)\n",
		       num, MAX_ALIST);
	Delete (in_or_out);
	if (opt_num > 0)
	    Delete (name);
	return (-1);
    }

    if (opt_num <= 0)
	name = "";
    config_notice (TR_PARSE, uii,
		 "CONFIG distribute-list %d %s %s\n", num, in_or_out, name);

    if (*name) {
	if ((interface = find_interface_byname (name)) == NULL) {
	    config_notice (TR_ERROR, uii,
			   "unkown interface %s\n", name);
	    Delete (in_or_out);
	    Delete (name);
	    return (-1);
	}
	if (strcasecmp (in_or_out, "in") == 0)
	    interface->dlist_in[CONFIG_MRTD->protocol] = num;
	else
	    interface->dlist_out[CONFIG_MRTD->protocol] = num;

	dlist = New (dlist_t);
	dlist->num = num;
	dlist->in_or_out = in_or_out;
	dlist->interface = interface;
	LL_Add (rip->ll_dlists, dlist);
	Delete (name);
    }
    else {
	LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	    if (strcasecmp (in_or_out, "in") == 0)
		interface->dlist_in[CONFIG_MRTD->protocol] = num;
	    else
		interface->dlist_out[CONFIG_MRTD->protocol] = num;
	}

	dlist = New (dlist_t);
	dlist->num = num;
	dlist->in_or_out = in_or_out;
	dlist->interface = NULL;
	LL_Add (rip->ll_dlists, dlist);
    }
    return (1);
}


static void
get_config_rt_rip (rip_t *rip)
{
    int i;
    prefix_t *prefix;
    interface_t *interface;
    dlist_t *dlist;

    config_add_output ("router %s\n", proto2string (rip->proto));

    LL_Iterate (rip->ll_networks, prefix) {
	config_add_output ("  network %s\n", prefix_toax (prefix));
    }

    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	if (BITX_TEST (&rip->interface_mask, interface->index))
	    config_add_output ("  network %s\n", interface->name);
    }

    LL_Iterate (rip->ll_dlists, dlist) {
        config_add_output ("  distribute-list %d %s %s\n",
                           dlist->num, dlist->in_or_out,
                           dlist->interface ? dlist->interface->name : "");
    }

    for (i = 0; i <= PROTO_MAX; i++) {
        if (BGP4_BIT_TEST (rip->redistribute_mask, i))
	    config_add_output ("  redistribute %s\n", proto2string (i));
    }
}


static void
get_config_router_rip (rip_t *rip)
{
    schedule_event_and_wait ("get_config_rt_rip", rip->schedule,
                             get_config_rt_rip, 1, rip);
}


#ifdef HAVE_IPV6
/*
 * start RIPNG protocol configuration 
 */
static int
config_router_ripng (uii_connection_t * uii)
{
    if (uii->negative) {
        if (!BGP4_BIT_TEST (MRT->protocols, PROTO_RIPNG))
	    return (0);
        BGP4_BIT_RESET (MRT->protocols, PROTO_RIPNG);
        config_del_module (CF_DELIM, "router ripng", get_config_router_rip, 
			   RIPNG);
        rip_stop (PROTO_RIPNG);
        return (1);
    }
    CONFIG_MRTD->protocol = PROTO_RIPNG;
    uii->state = UII_CONFIG_ROUTER_RIPNG;
    if (BGP4_BIT_TEST (MRT->protocols, PROTO_RIPNG))
	return (0);

    BGP4_BIT_SET (MRT->protocols, PROTO_RIPNG);
    config_add_module (CF_DELIM, "router ripng", get_config_router_rip, RIPNG);
    rip_start (PROTO_RIPNG);
    return (1);
}
#endif /* HAVE_IPV6 */


/*
 * start RIP protocol configuration 
 */
static int
config_router_rip (uii_connection_t * uii)
{
    if (uii->negative) {
        if (!BGP4_BIT_TEST (MRT->protocols, PROTO_RIP))
	    return (0);
        BGP4_BIT_RESET (MRT->protocols, PROTO_RIP);
        config_del_module (CF_DELIM, "router rip", get_config_router_rip, 
			   RIP);
        rip_stop (PROTO_RIP);
        return (1);
    }
    CONFIG_MRTD->protocol = PROTO_RIP;
    uii->state = UII_CONFIG_ROUTER_RIP;
    if (BGP4_BIT_TEST (MRT->protocols, PROTO_RIP))
	return (0);

    BGP4_BIT_SET (MRT->protocols, PROTO_RIP);
    config_add_module (CF_DELIM, "router rip", get_config_router_rip, RIP);
    rip_start (PROTO_RIP);
    return (1);
}


static int
config_router_rip_network_prefix (uii_connection_t * uii, prefix_t *prefix)
{
    interface_t *interface;
    LINKED_LIST *ll;
    rip_t *rip = RIP;
    prefix_t *network;

    assert (CONFIG_MRTD->protocol == PROTO_RIP ||
	CONFIG_MRTD->protocol == PROTO_RIPNG);

#ifdef HAVE_IPV6
    if (CONFIG_MRTD->protocol == PROTO_RIPNG)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    LL_Iterate (rip->ll_networks, network) {
        if (prefix_compare (prefix, network)) { 
	    break;
        }
    }

    if (uii->negative) {
	 if (network == NULL) {
	    Deref_Prefix (prefix);
	    return (0);
	}
	if (MRT->rib_redistribute_network)
            MRT->rib_redistribute_network (CONFIG_MRTD->protocol, prefix, 0);
        LL_Remove (rip->ll_networks, network);
	Deref_Prefix (network);
    }
    else {
        if (network) {
	    Deref_Prefix (prefix);
	    return (0);
	}
	if (MRT->rib_redistribute_network)
            MRT->rib_redistribute_network (CONFIG_MRTD->protocol, prefix, 1);
        LL_Add (rip->ll_networks, Ref_Prefix (prefix));
    }

    if ((ll = find_network (prefix)) != NULL) {
	LL_Iterate (ll, interface) {
#ifdef HAVE_IPV6
	    if (!uii->negative && rip->proto == PROTO_RIPNG) {
		if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
		    BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
			(interface->link_local == NULL ||
			 interface->link_local->broadcast == NULL)) {
			config_notice (TR_ERROR, uii,
				     "RIPNG on interface %s ignored "
			    "due to no multicast or link-local dest addr\n",
				       interface->name);
		    }
	    }
#endif /* HAVE_IPV6 */

		if (uii->negative) {
		    if (BITX_TEST (&rip->interface_mask, interface->index)) {
		        BITX_RESET (&rip->interface_mask, interface->index);
		        config_notice (TR_TRACE, uii, 
				       "%s interface %s (off)\n",
				       proto2string (rip->proto),
				       interface->name);
		        rip_interface (rip, interface, 0);
		    }
		}
		else {
		    if (!BITX_TEST (&rip->interface_mask, interface->index)) {
		        BITX_SET (&rip->interface_mask, interface->index);
		        /* interface->default_pref = RIP_PREF; */
		        config_notice (TR_TRACE, uii, 
				       "%s on interface %s (on)\n",
				       proto2string (rip->proto),
				       interface->name);
		        rip_interface (rip, interface, 1);
		    }
		}
	}
	LL_Destroy (ll);
	Deref_Prefix (prefix);
	return (1);
    }
    /* else {
	Deref_Prefix (prefix);
	return (-1);
    } */
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_rip_network_interface (uii_connection_t * uii, char *name)
{
    interface_t *interface;
    rip_t *rip = RIP;

    assert (CONFIG_MRTD->protocol == PROTO_RIP ||
	CONFIG_MRTD->protocol == PROTO_RIPNG);

#ifdef HAVE_IPV6
    if (CONFIG_MRTD->protocol == PROTO_RIPNG)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    if ((interface = find_interface_byname (name)) == NULL) {
	config_notice (TR_ERROR, uii,
		       "could not find interface %s\n", name);
	Delete (name);
	return (-1);
    }
    Delete (name);

    if (!uii->negative && BITX_TEST (&rip->interface_mask, interface->index))
	return (0);
    if (uii->negative && !BITX_TEST (&rip->interface_mask, interface->index))
	return (0);

#ifdef HAVE_IPV6
	if (!uii->negative && rip->proto == PROTO_RIPNG) {
	    if (!BITX_TEST (&rip->interface_mask, interface->index)) {
		if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
		    BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
		    (interface->link_local == NULL ||
		     interface->link_local->broadcast == NULL)) {
		    config_notice (TR_ERROR, uii,
				   "RIPNG on interface %s ignored "
			    "due to no multicast or link-local dest addr\n",
				   interface->name);
		    return (-1);
		}
		
	    }
	}
#endif /* HAVE_IPV6 */

    if (uii->negative) {
        BITX_RESET (&rip->interface_mask, interface->index);
        config_notice (TR_TRACE, uii,
		       "%s on interface %s (off)\n", proto2string (rip->proto),
		       interface->name);
        /* LL_Remove (rip->ll_interfaces, interface); */
        rip_interface (rip, interface, 0);
    }
    else {
        BITX_SET (&rip->interface_mask, interface->index);
        /* interface->default_pref = RIP_PREF; */
        config_notice (TR_TRACE, uii,
		       "%s on interface %s\n", proto2string (rip->proto),
		       interface->name);
        /* LL_Add (rip->ll_interfaces, interface); */
        rip_interface (rip, interface, 1);
    }
    return (1);
}


/*
 * Redistribute a given protocol into another protocol 
 */
static int
config_router_rip_redistribute (uii_connection_t * uii, char *proto_string)
{
    rip_t *rip = RIP;
    int proto;

    if ((proto = string2proto (proto_string)) < 0) {
        Delete (proto_string);
	return (-1);
    }

    if (CONFIG_MRTD->protocol == proto) {
	config_notice (TR_ERROR, uii,
		 "%s redistribute %s -- the same protocols!\n",
		  proto2string (CONFIG_MRTD->protocol), proto2string (proto));
	Delete (proto_string);
	return (-1);
    }

#ifdef HAVE_IPV6
    if (CONFIG_MRTD->protocol == PROTO_RIPNG)
	rip = RIPNG;
#endif /* HAVE_IPV6 */

    if (uii->negative) {
        if (!BGP4_BIT_TEST (rip->redistribute_mask, proto)) {
	    Delete (proto_string);
	    return (0);
	}
        BGP4_BIT_RESET (rip->redistribute_mask, proto);
	if (MRT->rib_redistribute_request)
	    MRT->rib_redistribute_request (rip->proto, proto, 0);
        Delete (proto_string);
        return (1);
    }

    if (BGP4_BIT_TEST (rip->redistribute_mask, proto)) {
	Delete (proto_string);
	return (0);
    }
    BGP4_BIT_SET (rip->redistribute_mask, proto);
    if (MRT->rib_redistribute_request)
	MRT->rib_redistribute_request (rip->proto, proto, 1);
    Delete (proto_string);
    return (1);
}


void
config_rip_init (void)
{
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_RIP, "Router RIP> ", 0);
#ifdef HAVE_IPV6
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_RIPNG, "Router RIPng> ", 0);
#endif /* HAVE_IPV6 */

    /* Interactive commands */
    uii_add_command_schedule (UII_NORMAL, COMMAND_NORM, "show rip", rip2_show,
		      "Show RIP status", RIP->schedule);
    uii_add_command_schedule (UII_NORMAL, COMMAND_NORM, "show rip routes",
		      rip2_show_routing_table, "Show RIP routing table",
		      RIP->schedule);

#ifdef HAVE_IPV6
    uii_add_command_schedule (UII_NORMAL, COMMAND_NORM, "show ripng", 
		      ripng_show,
		      "Show RIPng Status", RIPNG->schedule);
    uii_add_command_schedule (UII_NORMAL, COMMAND_NORM, "show ripng routes",
		      ripng_show_routing_table,
		      "Show RIPng routing table", RIPNG->schedule);
#endif /* HAVE_IPV6 */

    /* Config mode */
    uii_add_command_schedule (UII_CONFIG, COMMAND_NORM, "router rip",
		      config_router_rip, "Enables RIP routing protocol",
		      RIP->schedule);
    uii_add_command_schedule (UII_CONFIG, COMMAND_NORM, "no router rip",
		      config_router_rip, "Disables RIP routing protocol",
		      RIP->schedule);
#ifdef HAVE_IPV6
    uii_add_command_schedule (UII_CONFIG, COMMAND_NORM, "router ripng",
		      config_router_ripng, "Enables RIPng routing protocol",
		      RIPNG->schedule);
    uii_add_command_schedule (UII_CONFIG, COMMAND_NORM, "no router ripng",
		      config_router_ripng, "Disable RIPng routing protocol",
		      RIPNG->schedule);
#endif /* HAVE_IPV6 */

    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, 
		      "network %p",
		      config_router_rip_network_prefix, 
		      "Turns ON RIP routing for the network",
		      RIP->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, 
		      "network %n",
		      config_router_rip_network_interface, 
		      "Turns ON RIP routing on the interface",
		      RIP->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, 
		      "no network %p",
		      config_router_rip_network_prefix, 
		      "Turns OFF RIP routing for the network",
		      RIP->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, 
		      "no network %n",
		      config_router_rip_network_interface, 
		      "Turns OFF RIP routing on the interface",
		      RIP->schedule);
#ifdef HAVE_IPV6
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, 
		      "network %P",
		      config_router_rip_network_prefix, 
		      "Turns ON RIPng routing for the network",
		      RIPNG->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, 
		      "network %n",
		      config_router_rip_network_interface, 
		      "Turns ON RIPng routing on the interface",
		      RIPNG->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, 
		      "no network %P",
		      config_router_rip_network_prefix, 
		      "Turns OFF RIPng routing for the network",
		      RIPNG->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, 
		      "no network %n",
		      config_router_rip_network_interface, 
		      "Turns OFF RIPng routing on the interface",
		      RIPNG->schedule);
#endif /* HAVE_IPV6 */

    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
	"redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_router_rip_redistribute,
		      "Redistributes route from the protocol",
		      RIP->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
	"no redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_router_rip_redistribute,
		      "Not redistribute route from the protocol",
		      RIP->schedule);
#ifdef HAVE_IPV6
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
	"redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_router_rip_redistribute,
		      "Redistributes route from the protocol",
		      RIPNG->schedule);
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
	"no redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_router_rip_redistribute,
		      "Not redistribute route from the protocol",
		      RIPNG->schedule);
#endif /* HAVE_IPV6 */

    uii_add_command_schedule (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
		      "distribute-list %d (in|out) [%s]",
		      config_distribute_list,
		      "Applies access list to route [on interface]",
		      RIP->schedule);
#ifdef HAVE_IPV6
    uii_add_command_schedule (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
		      "distribute-list %d (in|out) [%s]",
		      config_distribute_list,
		      "Applies access list to route [on interface]",
		      RIPNG->schedule);
#endif /* HAVE_IPV6 */
}
